from dataclasses import dataclass, field
from math import ceil


@dataclass
class MessageSmartEncoding:
    """
    Sms Message Representation
    """

    replacements = {
        "“": '"',
        "”": '"',
        "‘": "'",
        "’": "'",
        "–": "-",
        "—": "-",
        "…": "...",
        "•": "-",
        "€": "EUR",
        "™": "",
        "©": "",
        "®": "",
        "\u0002": "",
        "\u001b": "",
    }
    body: str
    normalized_text: str = field(init=False)
    length: int = field(init=False)
    segments: int = field(init=False)

    def __post_init__(self):
        self.normalized_text = self.normalize(self.body)
        self.length = len(self.normalized_text)
        self.segments = ceil(self.length / 160)

    def normalize(self, text: str) -> str:
        for char, replacement in self.replacements.items():
            text = text.replace(char, replacement)
        return text



import json
import re
from datetime import datetime
from decimal import Decimal
from functools import cached_property
from typing import Dict

import boto3
from boto3.dynamodb.conditions import Key
from botocore.exceptions import ClientError

from reach_commons.app_logging.logger import get_reach_logger
from reach_commons.utils import DecimalEncoder, build_update_params_for_attributes


# noinspection PyMethodMayBeStatic
class BaseDynamoDBClient:
    def __init__(
        self,
        region_name="us-east-1",
        profile_name=None,
    ):
        self.region_name = region_name
        self.profile_name = profile_name


class DynamoDBClient(BaseDynamoDBClient):
    @cached_property
    def client(self):
        session = boto3.Session(
            region_name=self.region_name, profile_name=self.profile_name
        )
        return session.client("dynamodb")

    def put_item(self, table_name, item):
        return self.client.put_item(TableName=table_name, Item=item)


class SMSConversationsComonQueries:
    @staticmethod
    def clean_phone_number(phone_number):
        return re.sub(r"\D", "", phone_number)

    @staticmethod
    def add_message_to_conversation_v2(
        table,
        from_number,
        to_number,
        business_id,
        business_number,
        body,
    ):
        from_number = SMSConversationsComonQueries.clean_phone_number(from_number)
        to_number = SMSConversationsComonQueries.clean_phone_number(to_number)
        business_number = SMSConversationsComonQueries.clean_phone_number(
            business_number
        )
        other_number = from_number if from_number != business_number else to_number

        timestamp = datetime.utcnow().isoformat() + "Z"
        conversation_key = f"CHAT#{other_number}#TIMESTAMP#{timestamp}"

        status_message = "SENT" if from_number == business_number else "RECEIVED"

        item = {
            "PK": f"BUSINESS#{business_id}",
            "SK": conversation_key,
            "from_number": from_number,
            "to_number": to_number,
            "created_at": timestamp,
            "updated_at": timestamp,
            "status_message": status_message,
            "body": body,
        }

        table.put_item(Item=item)

        table.update_item(
            Key={"PK": f"BUSINESS#{business_id}", "SK": f"CONV#{other_number}"},
            UpdateExpression="SET last_message = :last_message, created_at = if_not_exists(created_at, :created_at)",
            ExpressionAttributeValues={
                ":last_message": item,
                ":created_at": timestamp,
            },
            ReturnValues="ALL_NEW",
        )

    @staticmethod
    def update_chat_status(table, pk, sk, status_message):
        timestamp = datetime.utcnow().isoformat() + "Z"
        table.update_item(
            Key={"PK": pk, "SK": sk},
            UpdateExpression="SET status_message = :status_message, updated_at = :updated_at",
            ConditionExpression="attribute_exists(PK) AND attribute_exists(SK)",
            ExpressionAttributeValues={
                ":status_message": status_message,
                ":updated_at": timestamp,
            },
            ReturnValues="ALL_NEW",
        )

    @staticmethod
    def add_phone_owner(table, phone_number, owner_name, additional_info):
        if not owner_name or not additional_info:
            get_reach_logger().warning(
                "Both 'owner_name' and 'additional_info' must be provided. Operation ignored."
            )
            return

        if not additional_info or not (
            "business_id" in additional_info or "customer_id" in additional_info
        ):
            raise ValueError(
                "additional_info must contain either 'business_id' or 'customer_id'."
            )

        phone_number = SMSConversationsComonQueries.clean_phone_number(phone_number)

        update_expression = "SET #name = :name"
        expression_attribute_names = {"#name": "name"}
        expression_attribute_values = {":name": owner_name}

        if additional_info:
            for key, value in additional_info.items():
                update_expression += f", #{key} = :{key}"
                expression_attribute_names[f"#{key}"] = key
                expression_attribute_values[f":{key}"] = value

        table.update_item(
            Key={"PK": f"OWNER#{phone_number}", "SK": "INFO"},
            UpdateExpression=update_expression,
            ExpressionAttributeNames=expression_attribute_names,
            ExpressionAttributeValues=expression_attribute_values,
        )

    @staticmethod
    def get_phone_owner(table, phone_number):
        phone_number = SMSConversationsComonQueries.clean_phone_number(phone_number)

        response = table.get_item(Key={"PK": f"OWNER#{phone_number}", "SK": "INFO"})
        return response.get("Item")


class BusinessReviewsCommonQueries:
    @staticmethod
    def get_review_business_info(table, business_id: str) -> Dict:
        query_params = {
            "KeyConditionExpression": Key("PK").eq(f"business#{business_id}")
            & Key("SK").eq("info")
        }

        response = table.query(**query_params)

        if response["Items"]:
            business_info = response["Items"][0]
            return business_info
        else:
            return {}

    @staticmethod
    def patch_business_info(
        table,
        business_id: str,
        business_info: Dict,
        created_at_isoformat: str = None,
    ) -> str:
        filtered_info = {k: v for k, v in business_info.items() if v is not None}
        filtered_info = json.loads(
            json.dumps(filtered_info, cls=DecimalEncoder), parse_float=Decimal
        )

        try:
            table.put_item(
                Item={
                    "PK": f"business#{business_id}",
                    "SK": "info",
                    "created_at": created_at_isoformat
                    if created_at_isoformat
                    else datetime.utcnow().isoformat(),
                    **filtered_info,
                },
                ConditionExpression="attribute_not_exists(PK) AND attribute_not_exists(SK)",
            )
            return "insert"
        except ClientError as e:
            if e.response["Error"]["Code"] == "ConditionalCheckFailedException":
                (
                    update_expr,
                    attr_values,
                    attr_names,
                ) = build_update_params_for_attributes(filtered_info)
                table.update_item(
                    Key={"PK": f"business#{business_id}", "SK": "info"},
                    UpdateExpression=update_expr,
                    ExpressionAttributeValues=attr_values,
                    ExpressionAttributeNames=attr_names,
                )
                return "update"
            else:
                raise



from urllib.parse import urlencode

from pymongo import InsertOne, MongoClient, UpdateOne
from pymongo.errors import BulkWriteError

from reach_commons.app_logging.logger import get_reach_logger


class MongoCustomerCRUD:
    def __init__(
        self,
        connection_info: dict,
        logger=get_reach_logger(),
    ):
        self.logger = logger
        self.mongo_client = MongoClient(
            "{}://{}:{}@{}/?{}".format(
                connection_info["schema"],
                connection_info["username"],
                connection_info["password"],
                connection_info["host"],
                urlencode(connection_info["extra_args"]),
            )
        )
        self.db = self.mongo_client.get_database(connection_info["database"])
        self.collection = self.db.get_collection("customers")

    def validate_and_prepare_operations(self, records: list, operation: str):
        operations = []
        for record in records:
            customer_id = record.get("customer_id")
            business_id = record.get("business_id")

            if not customer_id or not business_id:
                self.logger.error(
                    f"Missing customer_id or business_id in record. record={str(record)}"
                )
                raise ValueError(
                    f"Both customer_id and business_id are required in the record. record={str(record)}"
                )

            query = {"customer_id": customer_id, "business_id": business_id}

            if operation == "upsert":
                update_fields = {"$set": record}
                if "updated_at" not in record:
                    update_fields["$currentDate"] = {"updated_at": True}

                operations.append(UpdateOne(query, update_fields, upsert=True))
            elif operation == "insert":
                operations.append(InsertOne(record))
            elif operation == "update":
                update_fields = {"$set": record, "$currentDate": {"updated_at": True}}
                operations.append(UpdateOne(query, update_fields, upsert=False))
            elif operation == "delete":
                operations.append(UpdateOne(query, {"$set": {"deleted": True}}))
            else:
                raise ValueError(
                    f"Invalid operation: {operation}.  record={str(record)}"
                )

        return operations

    def execute_bulk_operations(self, operations: list):
        try:
            if operations:
                result = self.collection.bulk_write(operations)
                self.logger.info(f"Bulk operation completed: {result.bulk_api_result}")
        except BulkWriteError as bwe:
            self.logger.error(f"Bulk write error: {bwe.details}")

    def process_records(self, records: list, operation: str):
        operations = self.validate_and_prepare_operations(records, operation)
        self.execute_bulk_operations(operations)


"""
Uses example
connection_info = {
    "database": "db0",
    "schema": "mongodb+srv",
    "username": "databricks_rw",
    "password": "Lh4GsNx7cQUPtDmONqVGI9Dy7Jy6TbAxmq",
    "host": "cluster0.rwha2.mongodb.net",
    "extra_args": {
        "retryWrites": "true",
        "w": "majority",
        "appName": "Cluster0",
    },
}

mongo_crud = MongoCustomerCRUD(connection_info)

message = {
    "metadata": {"operation": "upsert"},
    "records": [
        {"customer_id": "123", "business_id": "456", "name": "John Doe"},
        {"customer_id": "124", "business_id": "456", "name": "Jane Doe"}
    ],
}


mongo_crud.process_records(message["records"], message["metadata"]["operation"])

"""



from functools import cached_property

import redis


class RedisManager:
    """
    # Example of usage
    # redis_manager = RedisManager('your_redis_host', your_redis_port, your_socket_timeout, your_socket_connection_timeout, your_max_retry)
    # redis_manager.add('key', 'value')
    # value = redis_manager.get('key')
    # redis_manager.delete('key')
    """

    def __init__(
        self,
        redis_host,
        redis_port,
        redis_socket_timeout=10,
        redis_socket_connection_timeout=10,
        redis_max_retry=10,
    ):
        self.redis_host = redis_host
        self.redis_port = redis_port
        self.redis_socket_timeout = redis_socket_timeout
        self.redis_socket_connection_timeout = redis_socket_connection_timeout
        self.redis_max_retry = redis_max_retry

    @cached_property
    def redis_connection(self):
        connection_pool = redis.ConnectionPool(
            host=self.redis_host,
            port=self.redis_port,
            socket_timeout=self.redis_socket_timeout,
            socket_connect_timeout=self.redis_socket_connection_timeout,
            max_connections=self.redis_max_retry,
        )
        try:
            redis_connection = redis.StrictRedis(connection_pool=connection_pool)
            ok = redis_connection.ping()
            if not ok:
                raise Exception("ERROR: Redis ping failed")
            return redis_connection
        except Exception as ex:
            raise ex

    def add(self, key, value, expire_seconds=None):
        if expire_seconds is not None:
            self.redis_connection.set(key, value, ex=expire_seconds)
        else:
            self.redis_connection.set(key, value)

    def get(self, key):
        return self.redis_connection.get(key)

    def delete(self, key):
        self.redis_connection.delete(key)

    def set(self, name, value, nx=False, ex=None):
        """
        Sets a key-value pair in Redis with optional constraints.

        Parameters:
        - name (str): The key name to set in Redis.
        - value (str): The value to store under the key.
        - nx (bool): If True, sets the key only if it does not already exist (used for locking).
        - ex (int | None): Expiration time in seconds. If provided, the key will automatically expire.

        Returns:
        - bool: True if the key was set successfully, False otherwise (especially when nx=True and key already exists).
        """
        return self.redis_connection.set(name=name, value=value, nx=nx, ex=ex)



from functools import cached_property
from io import BytesIO

import boto3

from reach_commons.app_logging.logger import get_reach_logger


class S3Client:
    def __init__(
        self,
        logger=get_reach_logger(),
        region_name="us-east-1",
        profile_name=None,
    ):
        self.logger = logger
        self.region_name = region_name
        self.profile_name = profile_name

    @cached_property
    def client(self):
        session = boto3.Session(
            region_name=self.region_name, profile_name=self.profile_name
        )

        return session.client("s3")

    def get_object(self, s3_bucket_name, s3_key):
        try:
            s3_object = self.client.get_object(Bucket=s3_bucket_name, Key=s3_key)
            actual_message_body = s3_object["Body"].read().decode("utf-8")

            self.logger.info(
                f"Retrieved object from S3: {s3_key} from bucket: {s3_bucket_name}"
            )
            return actual_message_body
        except Exception as e:
            self.logger.error(
                f"Error retrieving object {s3_key} from bucket: {s3_bucket_name}: {str(e)}"
            )

        return None

    def add_object(self, s3_bucket_name, s3_key, str_content):
        try:
            file_object = BytesIO(str_content.encode("utf-8"))

            self.client.upload_fileobj(
                Fileobj=file_object, Bucket=s3_bucket_name, Key=s3_key
            )
            url = (
                f"https://{s3_bucket_name}.s3.{self.region_name}.amazonaws.com/{s3_key}"
            )

            self.logger.info(
                f"Uploaded object to S3: {s3_key} in bucket: {s3_bucket_name}"
            )
            return url
        except Exception as e:
            self.logger.error(
                f"Error uploading object {s3_key} to bucket: {s3_bucket_name}: {str(e)}"
            )
            return None



import requests
from requests.auth import HTTPBasicAuth

from reach_commons.app_logging.logger import get_reach_logger


class CommonsPhoneNumberValidation:
    def __init__(self, logger=get_reach_logger()):
        self.base_url = "https://lookups.twilio.com/v1/PhoneNumbers/"
        self.account_sid = "AC1f9d2cb1c49cf9522f0de29861c9652d"
        self.auth_token = "2a5557dd0c80ea45bc23b3de68455480"
        self.logger = logger

    def is_valid_phone_number(self, phone_number):
        url = f"{self.base_url}{phone_number}"
        auth = HTTPBasicAuth(self.account_sid, self.auth_token)

        response = requests.get(url, auth=auth)

        if response.status_code == 200:
            return True
        else:
            self.logger.warning(
                f"Phone number {phone_number} is invalid. Status code: {response.status_code}, "
                f"Response: {response.text}"
            )
            return False



import copy
import datetime
import decimal
import functools
import json
import re
import warnings
import xml.etree.ElementTree as ET
from functools import reduce
from io import BytesIO
from sys import getsizeof
from typing import Any, Dict, Union

str_default_date = "%Y-%m-%dT%H:%M:%S"


class ReachDateUtils:
    @staticmethod
    def chunk_date_range(start_date, end_date, days_per_chunk):
        """
        Divides the interval between start_date and end_date into blocks of days_to_request days.
        Returns a list of tuples, each representing a date range.
        Each new range starts the day after the end of the previous range.
        """
        ranges = []
        current_start_date = start_date

        while current_start_date < end_date:
            current_end_date = current_start_date + datetime.timedelta(
                days=days_per_chunk - 1
            )

            # Ensure that the final interval does not exceed end_date
            if current_end_date > end_date:
                current_end_date = end_date
            ranges.append((current_start_date, current_end_date))

            # Updating current_start_date to the day after current_end_date
            current_start_date = current_end_date + datetime.timedelta(days=1)

        return ranges

    @staticmethod
    def convert_datetime_to_isoformat(obj: Union[Dict[str, Any], list]):
        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, datetime.datetime):
                    obj[key] = value.isoformat()
                elif isinstance(value, dict):
                    ReachDateUtils.convert_datetime_to_isoformat(value)
                elif isinstance(value, list):
                    for item in value:
                        ReachDateUtils.convert_datetime_to_isoformat(item)
        elif isinstance(obj, list):
            for index, item in enumerate(obj):
                if isinstance(item, datetime.datetime):
                    obj[index] = item.isoformat()
                elif isinstance(item, dict):
                    ReachDateUtils.convert_datetime_to_isoformat(item)
                elif isinstance(item, list):
                    ReachDateUtils.convert_datetime_to_isoformat(item)
        return obj

    @staticmethod
    def date_to_str(date: datetime, strf_format: str = None) -> str:
        return date.strftime(strf_format or str_default_date)

    @staticmethod
    def str_to_date(date: str, strf_format: str = None) -> datetime:
        return datetime.datetime.strptime(date, strf_format or str_default_date)


class ReachChunkObjectsUtils:
    MAX_FIREHOSE_SIZE = 900000

    @staticmethod
    def set_dict_value(obj, path, value):
        """

        :param obj: root object
        :param path:
        :param value:
        :return:
        """
        *path, last = path.split(".")
        for bit in path:
            obj = obj.setdefault(bit, {})
        obj[last] = value

    @staticmethod
    def chunk_list_items(
        obj_to_split: dict, path_of_array: str, max_size=MAX_FIREHOSE_SIZE
    ) -> list:
        """
        Function created to split a dict into smaller dict's, keeping the properties default.
        :param obj_to_split: root object to split.
        :param path_of_array: array path, separated by dots.
        :param max_size: maximum size in bytes to each
        :return: list
        """
        if path_of_array:
            map_list = path_of_array.split(".")
            edges = reduce(
                lambda d, key: d.get(key) if isinstance(d, dict) else None,
                map_list,
                obj_to_split,
            )
            if not edges:
                edges = [obj_to_split]
                path_of_array = None
        elif isinstance(obj_to_split, list):
            edges = obj_to_split
        else:
            raise Exception("path_of_array cannot be None for dict")
        list_of_items = []
        current_items = []
        for elem in edges:
            current_items.append(elem)

            if len(json.dumps(current_items).encode("utf-8")) >= max_size:
                current_edge = copy.deepcopy(obj_to_split)
                if path_of_array:
                    ReachChunkObjectsUtils.set_dict_value(
                        current_edge, path_of_array, current_items
                    )
                list_of_items.append(current_edge)
                current_items = []
        if current_items:
            current_edge = copy.deepcopy(obj_to_split)
            if path_of_array:
                ReachChunkObjectsUtils.set_dict_value(
                    current_edge, path_of_array, current_items
                )
            list_of_items.append(current_edge)
        return list_of_items

    @staticmethod
    def chunk_xml_by_tag(content: str, header: str, footer: str, tag_name):
        context = ET.iterparse(BytesIO(content.encode("utf-8")))
        list_xml = []

        for event, elem in context:
            if elem.tag in tag_name:
                list_xml.append(
                    f"{header}{ET.tostring(elem, encoding='unicode')}{footer}"
                )
        return list_xml

    @staticmethod
    def chunk_xml(content: str, header: str, footer: str, tag_name, max_size):
        context = ET.iterparse(BytesIO(content.encode("utf-8")))
        list_xml = []

        current_xml: str = ""
        for event, elem in context:
            if elem.tag in tag_name:
                current_xml += ET.tostring(elem, encoding="unicode")
            if getsizeof(current_xml) >= max_size:
                list_xml.append(f"{header}{current_xml}{footer}")
                current_xml = ""
        if not list_xml:
            list_xml.append(f"{header}{current_xml}{footer}")
            current_xml = ""
        if current_xml:
            list_xml.append(f"{header}{current_xml}{footer}")
        return list_xml

    @staticmethod
    def split_list_into_chunks(items, chunk_size):
        list_of_chunks = []
        for i in range(0, len(items), chunk_size):
            list_of_chunks.append(items[i : i + chunk_size])
        return list_of_chunks


def deprecated(reason):
    def decorator(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            warnings.simplefilter("always", DeprecationWarning)
            warnings.warn(
                f"{func.__name__} está obsoleta: {reason}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            warnings.simplefilter("default", DeprecationWarning)
            return func(*args, **kwargs)

        return new_func

    return decorator


def remove_nulls(data):
    if isinstance(data, dict):
        return {k: remove_nulls(v) for k, v in data.items() if v is not None}
    elif isinstance(data, list):
        return [remove_nulls(item) for item in data if item is not None]
    else:
        return data


def build_update_params_for_attributes(attributes_to_update):
    """
    Constructs the parameters needed for an update operation in DynamoDB, specifically
    for the provided attributes. It handles reserved words by using attribute name aliases.

    :param attributes_to_update: A dictionary containing the attribute names and their new values.
    :return: A tuple containing the UpdateExpression, ExpressionAttributeValues, and
             ExpressionAttributeNames for the update operation.
    """
    update_expression = "SET " + ", ".join(
        f"#{k} = :val{k}" for k in attributes_to_update.keys()
    )

    expression_attribute_values = {
        f":val{k}": v for k, v in attributes_to_update.items()
    }

    expression_attribute_names = {f"#{k}": k for k in attributes_to_update.keys()}

    return update_expression, expression_attribute_values, expression_attribute_names


class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return str(o)
        return super().default(o)


def is_valid_email(v):
    import re

    regex = r"^\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"
    return re.match(regex, v)


def clean_phone_number(phone_number):
    return re.sub(r"\D", "", phone_number)
